/*    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2013-
 *    Salvatore Ruggieri
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package dd.papers.ecmlpkdd2014.frechet;
import java.io.PrintWriter;
import java.util.TreeSet;

import rst.util.Cronos;
import rst.util.Distribution;
import rst.util.IOUtil;
import rst.util.NumberUtil;
import dd.util.ContingencyTable;
import dd.util.DDTable;

/** Frechet attacks. */
public class FrechetAttack {

	/** Test. */
	public static void main(String[] args) throws Exception 
	{
		int []supp1 = {20, 10, 5};
		int []proxy = {1, 5, 30};
		boolean first = true;
		for(int s : supp1) {
			boolean append = false;
			for(int p : proxy) { 
				attack("papers/ecmlpkdd2014/frechet", "credit", "personal_status=female_div_or_dep_or_mar", "class=bad", p, s, 10000, append, first);
				append = true;
				first = false;
			}
		}

		int []supp2 = {96, 48, 24};
		first = true;
		for(int s : supp2) {
			boolean append = false;
			for(int p : proxy) {
				attack("papers/ecmlpkdd2014/frechet", "adult_discretized", "race=NonWhite", "income=-50K", p, s, 10000, append, first);
				append = true;
				first = false;
			}
		}
	}

	/** Attack a dataset. */
	public static void attack(String wdir, String dataset, String PD, String dec, 
								int MAXPROXY, int MINSUPP, int TOPTABLES, boolean append, boolean first) throws Exception 
	{
		// set MAXPROXY to 0 to test all proxy related items

		Cronos c = new Cronos();
		c.start();
		DDTable tb = new DDTable();
		tb.setWorkingDir(wdir+"/"+dataset);
		tb.loadFromArff(dataset);
		tb.setDiscMetaData(PD, dec, MINSUPP);
		tb.closedItemset = true;
		tb.extractItemsets();
		c.stop();
		double elapsedFIM = c.elapsed();
		tb.initScan();
		int nt = 0;
		TreeSet<FrechetInfo> top = new TreeSet<FrechetInfo>( new FrechetInfo.CompLower() ); // keep to tables
		ContingencyTable tableLeft;
		Distribution distRD = new Distribution(-1, 1, 0.01);
		// find best inferred tables
		while( (tableLeft = tb.nextCT()) != null) {
			if( tableLeft.n1() == 0 || tableLeft.n2()==0 )
				continue; // if p1 or p2 are undefined
			nt++;
			int[] proxyItems = tb.getItemsNotIn(tableLeft);
			if( MAXPROXY > 0)
				NumberUtil.shuffle(proxyItems); // random shuffle
			ContingencyTable lower = new ContingencyTable(0, 0, 0, 0);
			int n = tableLeft.n();
			int n1 = tableLeft.n1();
			int n2 = tableLeft.n2();
			ContingencyTable upper = new ContingencyTable(n, n, n, n);
//			System.out.println("Left = " + tableLeft);
			int nrelated = 0;
			for(int related : proxyItems) {
				c.pause(); // we are now simulating the construction of background knowledge
				ContingencyTable tableCenter = tb.changePD(tableLeft, related);
				ContingencyTable tableRight = tb.changeDec(tableLeft, related);
				c.restart(); // restart clock
				// frechet inference
				frechetRefine(tableCenter, tableRight, lower, upper);
				if( ++nrelated == MAXPROXY)
					break;
			}
			double RDlower = lower.a/(double)n1 - upper.c/(double)n2;
			double RDupper = upper.a/(double)n1 - lower.c/(double)n2;
			double actual = tableLeft.rd();
//			System.out.println("" + RDlower  + "\t" + RDupper + "\t" + tableLeft.rd());
			distRD.add(actual); // distribution of actual
			FrechetInfo fi = new FrechetInfo(tableLeft, RDlower, RDupper, actual);
			top.add(fi);
			// keep only best tables
			if( nt > TOPTABLES )
				top.remove(top.first()); // best inferred bounds
		}
		tb.endScan();
		System.out.println("no. contingency tables: " + nt);
		String outfile = wdir+"/plots/frechet_" + dataset + "_" + MINSUPP;
		PrintWriter out = IOUtil.getWriter(outfile+ ".data", append);
		// output best inferred tables and their distribution of RD lower bound
		int rank=0;
		Distribution distTot = new Distribution(-1, 1, 0.01);
		Distribution distZero = new Distribution(-1, 1, 0.01);
		for(FrechetInfo fi : top.descendingSet()) {
			++rank;
			out.println("" + MAXPROXY + "\t" + rank + "\t" + fi.lower  + "\t" + fi.upper + "\t" + fi.actual
					+ "\t" + fi.unknownTable);
			distTot.add(fi.lower);
			if(fi.upper == fi.lower)
				distZero.add(fi.lower); // exact RD value inferred
		}
		out.close();
		out = IOUtil.getWriter(outfile+ ".aggr", append); // aggregate values
		int tot = 0;
		int zero = 0;
		int overall = 0;
	    for(int i=distTot.count.length-1; i>=0;--i) {
	    	int n = distTot.count[i];
	    	if( n == 0 ) 
	    		continue;
    		double v = i*distTot.step+distTot.min;
    		tot += n;
	    	zero += distZero.count[i];
	    	overall += distRD.count[i];
			out.println("" + MAXPROXY + "\t" + v + "\t" + tot  + "\t" + zero  + "\t" + overall);
	    }
	    out.close();
		c.stop();
		double elapsedDD = c.elapsed();
		outfile = wdir+"/plots/frechet_" + dataset + "_time";
		out = IOUtil.getWriter(outfile+ ".data", !first);
		if(!append) {
			if(!first)
				out.println();
			out.print(MINSUPP + "\t" + TOPTABLES + "\t" + elapsedFIM);
		}
		out.print("\t" + MAXPROXY + "\t" + elapsedDD);
		
	    out.close();
		System.out.println("Elapsed time (s): " + c.elapsed());
	}

	/** Refine Frechet bounds. */
	private static void frechetRefine(ContingencyTable tableCenter, ContingencyTable tableRight,
				ContingencyTable lower, ContingencyTable upper) {
		
		int upper_a = Math.min(tableCenter.a, tableRight.a) + Math.min(tableCenter.c, tableRight.b);
		int lower_a = Math.max(tableRight.a-tableCenter.b, 0) + Math.max(tableRight.b-tableCenter.d, 0);
		int n1 = tableRight.n1();
		upper.a = Math.min(upper.a, upper_a);
		lower.b = n1 - upper.a;
		lower.a = Math.max(lower.a, lower_a);
		upper.b = n1 - lower.a;	
	
		int upper_c = Math.min(tableCenter.a, tableRight.c) + Math.min(tableCenter.c, tableRight.d);
		int lower_c = Math.max(tableRight.c-tableCenter.b, 0) + Math.max(tableRight.d-tableCenter.d, 0);
		int n2 = tableRight.n2();
		upper.c = Math.min(upper.c, upper_c);
		lower.d = n2 - upper.c;
		lower.c = Math.max(lower.c, lower_c);
		upper.d = n2 - lower.c;	
	}
}
